setClass("polynomial", representation(coef = "numeric", rat = "logical"),
         validity = function(object) {
           coef <- object@coef
           rat <- object@rat
           if(length(coef) > 1 &&
              abs(coef[length(coef)]) < .Machine$double.eps)
             return("zero leading coefficient")
           else if(length(rat) > 1 || !is(rat, "logical"))
             return("improper rationality marker")
           else return(TRUE)
         })

.clip <- function(p) {
  p <- as(p, "numeric")
  if((j <- length(p)) == 0) return(0)
  while(j > 1 && p[j] == 0) j <- j - 1
  p[1:j]
}

polynomial <- function(coef = c(0, 1), rat = all(coef %% 1 == 0))
  new("polynomial", coef = .clip(coef), rat = rat)

setAs("polynomial", "numeric", function(object) object@coef)

setMethod("Math", "polynomial",
          function(x) {
            switch(.Generic,
                   ceiling =, floor =,
                   trunc = polynomial(callGeneric(x@coef), rat = x@rat),
                   stop(paste(.Generic, "not allowed on polynomials"))
                   )}
          )

setMethod("Math2", "polynomial",
          function(x, digits)
          polynomial(callGeneric(x@coef, digits), rat = x@rat)
         )

setMethod("[", "polynomial", function(x, ..., drop = T)
          polynomial(x@coef[..., drop = drop], rat = x@rat))

setReplaceMethod("[", signature(x = "polynomial"),
                 function(x, ..., value) {
                   cx <- as(x, "numeric")
                   cx[...] <- as(value, "numeric")
                   polynomial(cx, rat = x@rat)
                 })

.poly.mult <- function(e1, e2) {
  .C("VR_polymul",
     e1, length(e1), e2, length(e2),
     e1.e2 = double(length(e1) + length(e2) - 1),
     COPY = rep(F, 5),
     CLASSES = c("numeric", "integer",
       "numeric", "integer", "numeric"))$e1.e2
}

.poly.quo.rem <- function(e1, e2) {
  l1 <- length(e1)
  l2 <- length(e2)
  if(l1 < l2)
    return(list(quotient = 0, remainder = e1))
  if (l2 == 1)
    return(list(quotient = e1/e2, remainder = 0))
  .C("VR_polydiv", remainder = e1, l1, e2, l2,
     quotient = double(l1 - l2 + 1),
     COPY = c(T, F, F, F, F),
     CLASSES = c("numeric", "integer",
       "numeric", "integer", "numeric"))[c("quotient", "remainder")]
}

.ArithPolynomial <- function(e1, e2) {
  rat <- (if(is(e1, "polynomial")) e1@rat else all(e1 %% 1 == 0)) &&
         (if(is(e2, "polynomial")) e2@rat else all(e2 %% 1 == 0))
  e1 <- as(e1, "numeric")
  e2 <- as(e2, "numeric")
  e1.op.e2 <- switch(.Generic,
    "+" = ,
    "-" = {
      if((j <- length(e1) - length(e2)) < 0) e1 <- c(e1, rep(0, -j))
      else if(j > 0) e2 <- c(e2, rep(0, j))
      callGeneric(e1, e2)
    },
    "*" = .poly.mult(e1, e2),
    "/" = ,
    "%/%" = .poly.quo.rem(e1, e2)$quotient,
    "%%" = .poly.quo.rem(e1, e2)$remainder,
    "^" = if(length(e2) > 1 || e2 < 0 || (e2 %% 1) != 0) {
      stop("positive integer powers only")
    } else {
      switch(as.character(e2),
             "0" = 1, "1" = e1,
             {
               m <- e1
               for(i in 2:e2) m <- .poly.mult(m, e1)
               m
             })
    })
  polynomial(e1.op.e2, rat = rat)
}

setMethod("Arith", signature(e1 = "polynomial", e2 = "missing"),
          function(e1, e2) {
            rat <- e1@rat
            e1 <- e1@coef
            polynomial(callGeneric(), rat = rat)
          })

setMethod("Arith", signature(e1 = "polynomial"), .ArithPolynomial)
setMethod("Arith", signature(e2 = "polynomial"), .ArithPolynomial)

.ComparePolynomial <- function(e1, e2) {
  e1 <- as(e1, "numeric")
  e2 <- as(e2, "numeric")
  switch(.Generic,
         "==" = length(e1) == length(e2) && all(e1 == e2),
         "!=" = length(e1) != length(e2) || any(e1 != e2),
         stop("unsupported comparison of polynomials"))
}

setMethod("Compare", signature(e1 = "polynomial"), .ComparePolynomial)
setMethod("Compare", signature(e2 = "polynomial"), .ComparePolynomial)

setMethod("Logic", signature(e1 = "polynomial"), function(e1, e2)
          stop("logical operations not allowed on polynomials"))
setMethod("Logic", signature(e2 = "polynomial"), function(e1, e2)
          stop("logical operations not allowed on polynomials"))

setIs("polynomial", "function", coerce =
      function (object) {
        p <- as.name("p")
        x <- as.name("x")
        object <- as(object, "numeric")
        if ((an <- length(object)) == 1) {
          object <- c(object, 0)
          an <- 2
        }
        statement <- expression()
        statement[[i <- 1]] <-
          structure(list(p, call("+", object[an-1],
                                 call("*", x, object[an]))), mode = "<-")
        for (ai in rev(object)[-(1:2)])
          statement[[i <- i + 1]] <-
            structure(list(p, call("+", ai, call("*", x, p))), mode = "<-")
        statement[[i]] <- statement[[i]][[2]]
        statement <-
          if(an == 2) statement[[1]]
          else as(statement, "{")
        substitute(function(x) body, list(body = statement))
      })

calcCoef <- function(x, y) {
  if(missing(y)) {
    ## Polynomial with specified zeros
    p <- 1
    for(xi in x) p <- c(0, p) - c(xi * p, 0)
    return(p)
  }
  ## Polynomial with specified values
  ## (Lagrange interpolation formula)
  if((length(x)) != length(y)) stop("length(x) != length(y)")
  if(any(discard <- duplicated(x))) {
    y <- y[!discard]
    x <- x[!discard]
  }
  if((m <- length(x)) <= 1) return(y)
  r <- 0
  for(i in 1:m)
    r <- r + (y[i] * Recall(x[ - i]))/prod(x[i] - x[ - i])
  r
}

polyWithZeros <- polyWithRoots <- function(z, rat = F)
  polynomial(calcCoef(z), rat = rat)

polyInterp <- function(x, y, rat = F)
  polynomial(calcCoef(unclass(x), unclass(y)), rat = rat)

setAs("polynomial", "character",
      function(object) {
        p <- object@coef
        tap <- if(object@rat) .vfrac(abs(p)) else
        as(zapsmall(abs(p)), "character")
        lp <- length(p) - 1
        names(p) <- 0:lp
        if(any(discard <- match(tap, c("0", "-0"), nomatch = F))) {
          p <- p[!discard]
          if(length(p) == 0) return("0")
          tap <- tap[!discard]
        }
        sp <- c("- ", "+ ")[(p > 0) + 1]
        np <- names(p)
        p <- tap
        p[p == "1" & np != "0"] <- ""
        pow <- paste("x", "^", np, sep = "")
        pow[np == "0"] <- ""
        pow[np == "1"] <- "x"
        stars <- rep("*", length(p))
        stars[p == "" | pow == ""] <- ""
        p <- paste(sp, p, stars, pow, sep = "", collapse = " ")
        while((ch <- substring(p, 1, 1)) == "+" || ch == " ")
          p <- substring(p, 2)
        p
      })

print.polynomial <- function(x, ...) {
  p <- as(x, "character")
  ow <- max(35, options("width")$width - 3)
  m2 <- 0
  np <- nchar(p)
  while(m2 < np) {
    m1 <- m2 + 1
    m2 <- m2 + ow
    if(m2 < np)
      while(m2 > m1 && substring(p, m2, m2) != " ") m2 <- m2 - 1
    cat(substring(p, m1, m2), "\n")
  }
  invisible(x)
}

setMethod("show", "polynomial",
          function(object) print.polynomial(object))

solve.polynomial <- function(a, b, ...)
  sort(if(missing(b))
       polyroot(a@coef) else polyroot((a-b)@coef))

if(exists("integral", where = 1) && isGeneric("integral"))
  invisible(removeGeneric("integral"))

setGeneric("integral", function(object, limits = NULL, ...)
           standardGeneric("integral"))

setMethod("integral", signature(object = "polynomial"),
          function (object, limits = NULL, ...) {
            expr <- object@coef
            p <- polynomial(c(0, expr/seq(along = expr)), rat = object@rat)
            if (is.null(limits)) p
            else diff(predict(p, limits))
          })

coef.polynomial <- function (object, ...) object@coef

summary.polynomial <- function(object) {
  dp <- deriv(object)
  r <- structure(list(zeros = solve(object),
                      stationaryPoints = solve(dp),
                      inflexionPoints = solve(deriv(dp))),
                 originalPolynomial = object)
  oldClass(r) <- "polySummary"
  r
}

print.polySummary <- function(x, ...) {
  cat("\n Summary information for polynomial:\n")
  print(attr(x, "originalPolynomial"))
  cat("\n Zeros:\n")
  print(x$zeros)
  cat("\n Stationary points:\n")
  print(x$stationaryPoints)
  cat("\n Points of inflexion:\n")
  print(x$inflexionPoints)
  invisible(x)
}

predict.polynomial <- function (object, newdata, ...) {
  v <- 0
  object <- rev(object@coef)
  for (pj in object) v <- newdata * v + pj
  v
}

deriv.polynomial <- function (x, ...) {
  cx <- x@coef
  if (length(cx) == 1)
    return(polynomial(0))
  cx <- cx[-1]
  polynomial(cx * seq(along = cx), x@rat)
}

setMethod("xyCall", signature("polynomial", "missing"),
          function(x, y, FUN, ..., xexpr, yexpr) {
            pu <- par("usr")
            x0 <- seq(pu[1], pu[2], len = 300)
            y0 <- predict(x, x0)
            y0[y0 <= pu[3] | y0 > pu[4]] <- NA
            FUN(x0, y0, ...)
          })

plot.polynomial <- function (p, xlim = 0:1, ylim = range(Px),
                             type = "l", len = 300, ...) {
  if (missing(xlim))
    xlim <- range(Re(unlist(summary(p))))
  if (any(is.na(xlim))) {
    warning("summary of polynomial fails. Using nominal xlim")
    xlim <- 0:1
  }
  if (diff(xlim) == 0)
    xlim <- xlim + c(-1, 1)/2
  if (length(xlim) > 2) ### xlim gives the points precisely
    x <- xlim
  else {
    eps <- diff(xlim)/100
    xlim <- xlim + c(-eps, eps)
    x <- seq(xlim[1], xlim[2], len = len)
  }
  Px <- predict(p, x)
  if (!missing(ylim))
    Px[Px < ylim[1]] <- Px[Px > ylim[2]] <- NA
  plot(x, Px, type = type, xlim = xlim, ylim = ylim, ...)
}

.vfrac <- function(x, ...) {
  ans <- .rat(x, ...)
  ndc <- paste(ans$rat[, 1], ans$rat[, 2], sep = "/")
  int <- ans$rat[, 2] == 1
  ndc[int] <- as.character(ans$rat[int, 1])
  ndc
}

.rat <- function(x, cycles = 10, max.denominator = 2000) {
  a0 <- rep(0, length(x))
  A <- matrix(b0 <- rep(1, length(x)))
  B <- matrix(floor(x))
  r <- as.vector(x) - drop(B)
  len <- 0
  while(any(which <- (r > 1/max.denominator)) &&
        (len <- len + 1) <= cycles) {
    a <- a0
    b <- b0
    a[which] <- 1
    r[which] <- 1/r[which]
    b[which] <- floor(r[which])
    r[which] <- r[which] - b[which]
    A <- cbind(A, a)
    B <- cbind(B, b)
  }
  pq1 <- cbind(b0, a0)
  pq <- cbind(B[, 1], b0)
  len <- 1
  while((len <- len + 1) <= ncol(B)) {
    pq0 <- pq1
    pq1 <- pq
    pq <- B[, len] * pq1 + A[, len] * pq0
  }
  list(rat = pq, x = x)
}

poly.orth <- function(x, degree = length(unique(x)) - 1, norm = T)
{
  at <- attr(poly(x, degree), "coefs")
  a <- at$alpha
  N <- at$norm2
  x <- polynomial()
  p <- list(polynomial(0), polynomial(1))
  for(j in 1:degree)
    p[[j + 2]] <- (x - a[j]) * p[[j + 1]] - N[j + 1]/N[j] * p[[j]]
  p <- p[-1]
  if(norm) {
    sqrtN <- sqrt(N[-1])
    for(j in 1 + 0:degree) p[[j]] <- p[[j]]/sqrtN[j]
  }
  names(p) <- paste("degree", 0:degree)
  p
}
